import random


class Agent:
    
    def __init__(self, game):
        self.game = game
    
    def decision(self, state):
        pass
    
    def game_loop(self):
        
        state = self.game.start_state
        
        while True:
            
            action = self.decision(state)
            state = self.game.get_successor(state, action)
            self.game.print_state(state)
            if self.game.game_over(state):
                break
            
            row = int(input())
            col = int(input())
            action = (row, col, 'O')
            state = self.game.get_successor(state, action)
            self.game.print_state(state)
            if self.game.game_over(state):
                break
        
        if self.game.has_won(state, 'X'):
            print('Agent wins!')
        elif self.game.has_won(state, 'O'):
            print('Human wins!')
        else:
            print('Draw!')


class RandomAgent(Agent):
    
    def __init__(self, game):
        super().__init__(game)
    
    def decision(self, state):
        actions = self.game.get_actions(state)
        return random.choice(actions)


class MinMaxAgent(Agent):
    
    def __init__(self, game, depth_limit):
        super().__init__(game)
        self.depth_limit = depth_limit
        
    def decision(self, state):
        
        best_v = float('-inf')
        best_a = None
        
        actions = self.game.get_actions(state)
        for action in actions:
            next_v = self.min_value(self.game.get_successor(state, action), 0)
            if best_v < next_v:
                best_v = next_v
                best_a = action
        
        return best_a
    
    def max_value(self, state, depth):
        if self.game.game_over(state) or depth == self.depth_limit:
            return self.game.get_utility(state)
        actions = self.game.get_actions(state)
        v = float('-inf')
        for action in actions:
            next_v = self.min_value(self.game.get_successor(state, action), depth)
            v = max(v, next_v)
        return v
            
    def min_value(self, state, depth):
        if self.game.game_over(state) or depth == self.depth_limit:
            return self.game.get_utility(state)
        actions = self.game.get_actions(state)
        v = float('+inf')
        for action in actions:
            next_v = self.max_value(self.game.get_successor(state, action), depth + 1)
            v = min(v, next_v)
        return v