from copy import deepcopy


class XOGame:
    
    def __init__(self):
        self.start_state = {
            'board': [['_' for j in range(3)] for i in range(3)],
            'player': 'X'
        }
    
    def get_actions(self, state):
        actions = []
        for i in range(3):
            for j in range(3):
                if state['board'][i][j] == '_':
                    actions.append((i, j, state['player']))
        return actions
    
    def get_successor(self, state, action):
        new_state = deepcopy(state)
        i, j, player = action
        new_state['board'][i][j] = player
        new_state['player'] = 'O' if player == 'X' else 'X'
        return new_state
    
    def has_won(self, state, player):
        board = state['board']
        for i in range(3):
            if all([board[i][j] == player for j in range(3)]):
                return True
            if all([board[j][i] == player for j in range(3)]):
                return True
        if all([board[i][i] == player for i in range(3)]):
            return True
        if all([board[i][2-i] == player for i in range(3)]):
            return True
    
    def game_over(self, state):
        if self.has_won(state, 'X') or self.has_won(state, 'O'):
            return True
        for i in range(3):
            for j in range(3):
                if state['board'][i][j] == '_':
                    return False
        return True
    
    def get_utility(self, state):
        if self.has_won(state, 'X'):
            return float('+inf')
        elif self.has_won(state, 'O'):
            return float('-inf')
        elif self.game_over(state):
            return 0
    
        open_x = 8
        open_o = 8
        board = state['board']
        for i in range(3):
            
            if any([board[i][j] == 'O' for j in range(3)]):
                open_x -= 1
            if any([board[i][j] == 'X' for j in range(3)]):
                open_o -= 1
            
            if any([board[j][i] == 'O' for j in range(3)]):
                open_x -= 1
            if any([board[j][i] == 'X' for j in range(3)]):
                open_o -= 1
            
        if any([board[i][i] == 'O' for i in range(3)]):
            open_x -= 1
        if any([board[i][i] == 'X' for i in range(3)]):
            open_o -= 1
            
        if any([board[i][2-i] == 'O' for i in range(3)]):
            open_x -= 1
        if any([board[i][2-i] == 'X' for i in range(3)]):
            open_o -= 1
        
        return open_x - open_o
    
    def print_state(self, state):
        for row in state['board']:
            print('|'.join(row))
        print()